"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""


#!/usr/bin/python
# Aerospike python library for AMC

from constants import MAX_CONNECTIONS_PER_NODE
import socket                    # socket needs no introduction
import struct                    # gives us a parser/encoder for binary data
import types
import Queue

TIMEOUT = 0.5

#
# COMPATIBILITY COMPATIBILITY COMPATIBILITY
#
# So the 'struct' class went through lots of (good) improvements in 2.5, but we want to support
# old use as well as new. Write a few functions similar to the 2.5 ones, and either use builtin or
# pure based on what's available
#
#

def my_unpack_from(fmtStr, buf, offset  ):
    sz = struct.calcsize(fmtStr)
    return struct.unpack(fmtStr, buf[offset:offset+sz])

def my_pack_into(fmtStr, buf, offset, *args):
    tmp_array = struct.pack(fmtStr, *args)
    buf[offset:offset+len(tmp_array)] = tmp_array

# 2.5+ has this nice partition call
def partition_25(s, sep):
    return( s.partition(sep) )

# 2.4- doesn't
def partition_old(s, sep):
    idx = s.find(sep)
    if idx == -1:
        return(s, "", "")
    return( s[:idx], sep, s[idx+1:] )


g_proto_header = None
g_struct_header_in = None
g_struct_header_out = None
g_partition = None

# 2.5, this will succeed
try:
    g_proto_header = struct.Struct( '! Q' )
    g_struct_header_in = struct.Struct( '! Q B 4x B I 8x H H' )
    g_struct_header_out = struct.Struct( '! Q B B B B B B I I I H H' )
    g_partition = partition_25

# pre 2.5, if there's no Struct submember, so use my workaround pack/unpack
except:
    struct.unpack_from = my_unpack_from
    struct.pack_into = my_pack_into
    g_partition = partition_old


class SocketManager(object):

    def __init__(self):
        self.cache = {}

    def get_socket(self, *address):

        if address in self.cache:
            q = self.cache[address]['pool']
        else:
            q = Queue.Queue(maxsize=MAX_CONNECTIONS_PER_NODE)
            self.cache[address] = dict(pool=q, active=0)

        if q.empty() is False or self.cache[address]['active'] >= MAX_CONNECTIONS_PER_NODE:

            # pick available socket from pool or
            # if 10 active sockets with the specific node are already open,
            # keep waiting until one of these is freed to the pool and
            # then pick it up

            sock = q.get()
        else:

            # if no socket is available in the pool, and less than 10
            # have been already opened, create one

            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(TIMEOUT)
            try:
                sock.connect(address)
            except:
                sock.close()

                # retry the request with another socket to avoid failure
                # due to idle-timeout of server-side socket

                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(TIMEOUT)
                try:
                    sock.connect(address)
                except:
                    sock.close()
                    print "retried with a new socket, but failed to connect"
                    return -1
            self.cache[address]['active'] += 1
        return sock

    def release_socket(self, sock, *address):

        self.cache[address]['pool'].put(sock)

    def close_all_sockets_by_address(self, *address):

        while not self.cache[address]['pool'].empty():
            sock = self.cache[address]['pool'].get()
            sock.close()
        del(self.cache[address])

    def close_all_sockets(self):

        for d in list(self.cache.values()):
            for sock in d['pool']:
                sock.close()
        self.cache = {}

    def close_a_socket(self, sock, *address):

        sock.close()
        self.cache[address]['active'] -= 1


sm = SocketManager()


def citrusleaf_info_request(host, port, buf, debug=False):

    # request over TCP
    try:
        sock = sm.get_socket(host, int(port))
        if sock == -1:
            return -1
        try:
            sock.send(buf)
        except:
            sm.close_a_socket(sock, host, int(port))

            # retry the request with another socket to avoid failure
            # due to idle-timeout of server-side socket

            sock = sm.get_socket(host, int(port))
            try:
                sock.send(buf)
            except:
                sm.close_a_socket(sock, host, int(port))
                print "retried with a new socket, but failed to send"
                return -1

        if debug:
            print "info get response"
        # get response
        rsp_hdr = sock.recv(8)
        if debug:
            print "response is: "
            myHexlify(rsp_hdr)
        q = struct.unpack_from("! Q",rsp_hdr, 0)
        sz = q[0] & 0xFFFFFFFFFFFF
        if debug:
            print "recv header length ",sz
        if sz > 0:
            pos = 0
            while (pos < sz):
                tmp_data = sock.recv(sz - pos)
                if pos == 0:
                    rsp_data = tmp_data
                else:
                    rsp_data += tmp_data
                pos += len(tmp_data)
            if debug:
                print "recv body"
                myHexlify(rsp_data)
        sm.release_socket(sock, host, int(port))
    except Exception, ex:
        print "info request got exception ",type(ex)," ",ex
        return -1

    # parse out responses
    if sz == 0:
        return None

    if debug:
        print "receive as string: ",rsp_data

    return( rsp_data )



# Citrusleaf Info request
#
# This is a special purpose request to get informational name-value pairs
# from a given node. It's good for discovering the rest of the cluster,
# or trying to figure out which cluster members have which parts of the key space
#
# host, port are self explanatory
# 'names' is an iterable list of values to get, or None to get all
#  being really nice, also supporting a single string as a name instead of requiring a list
#
# Returns: values, a dict of with the name being the index, and the value being the value


def citrusleaf_info( host, port, names=None, debug=False):

    # Passed a set of names: created output buffer

    if names == None:
        q = (2 << 56) | (1 << 48)
        if g_proto_header != None:
            buf = g_proto_header.pack(q)
        else:
            buf = struct.pack('! Q',q)

    elif type(names) == types.StringType:
        q = (2 << 56) | (1 << 48) | (len(names) + 1)
        fmtStr = "! Q %ds B" % len(names)
        buf = struct.pack(fmtStr, q, names, 10 )

    else: # better be iterable of strings
        # annoyingly, join won't post-pend a seperator. So make a new list
        # with all the seps in
        names_l = []
        for name in names:
            names_l.append(name)
            names_l.append("\n")
        namestr = "".join(names_l)
        q = (2 << 56) | (1 << 48) | (len(namestr))
        fmtStr = "! Q %ds" % len(namestr)
        buf = struct.pack(fmtStr, q, namestr )


    if debug:
        print "request buffer: "
        myHexlify(buf)

    rsp_data = citrusleaf_info_request( host, port, buf, debug )

    if rsp_data == -1:
        return -1

    # if the original request was a single string, return a single string
    if type(names) == types.StringType:

        lines = rsp_data.split("\n")
        name, sep, value = g_partition(lines[0],"\t")

        if name != names:
            print " problem: requested name ",names," got name ",name
            return(-1)
        return value

    else:
        rdict = dict()
        for line in rsp_data.split("\n"):
            if len(line) < 1:        # this accounts for the trailing '\n' - cheaper than chomp
                continue
            if debug:
                print " found line ",line
            name, sep, value = g_partition(line,"\t")
            if debug:
                print "    name: ",name," value: ",value
            rdict[name] = value

        return rdict

# there is probably a cooler way to do this using some kind of reduce
# function. One where the lambda would return ord(c) or ord(c) + '\n'
#
		
def myHexlify(buf):
	print "my hexlify: length ",len(buf)
	for i, c in enumerate(buf):
		print "%02x " % ord(c) ,
		if i % 16 == 15:
			print ""
		if i % 16 == 7:
			print ": ",	
	print
		
