"""
Licensed to the Apache Software Foundation (ASF) under one
or more contributor license agreements.  See the NOTICE file
distributed with this work for additional information
regarding copyright ownership.  The ASF licenses this file
to you under the Apache License, Version 2.0 (the
"License"); you may not use this file except in compliance
with the License.  You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

# -*- coding: utf-8 -*-
import uuid, sys, os.path
from node import Node
from libs import base
from constants import *
from datetime import datetime
from helpers.seed_down_handler import HandleSeedDown
from helpers.seed_down_handler_with_version import HandleSeedDownWithVersion
from helpers import ip_checker


class Cluster(object):
    """Cluster contains all information related to cluster"""

    def __new__(cls, ip, port):

        if base.is_node_available(ip, port) is False:
            return UNABLE_TO_CONNECT
        return super(Cluster, cls).__new__(cls, ip, port)

    def __init__(self, ip, port):

        self._ip = ip
        self._port = port
        self._seed = "%s:%s" % (ip, port)
        self._cluster_id = str(uuid.uuid4())

        self._nodes = {self._seed: Node(self._seed, self._cluster_id)}
        self._last_access_time = datetime.now()

        self._id = None
        self._node_list = set([self._seed])
        self._current_nodes = []
        self._index_list = {}
        self._node_visibilties = {}

        self.get_updates()
        self.set_status_on()
        self._latest_build_no = self.get_build_no_of_cluster_nodes().get("latest_build_no")
        sys.stderr.write("Aerospike Version for cluster with seed node %s: %s\n" % (self._ip, str(self._latest_build_no)))
        self._namespaces = self.get_namespaces()
        self._cluster_key = self._nodes[self._seed].get_cluster_key()

    def get_cluster_id(self):

        return self._cluster_id

    def get_updates(self, alerts_flag=False):

        enterprise_flag = os.path.isfile(ENTERPRISE_FLAG_FILE)
        node_list = set()
        if enterprise_flag and alerts_flag:
            cluster_integrity = "true"
            is_cluster_size_same = True
            node_cluster_size = None

        for node in list(self._nodes.values()):
            if not node.is_valid():
                continue
            node_list = node_list.union(node.get_node_list())
            if enterprise_flag and alerts_flag:
                if node.get_statistic("cluster_integrity") == "false":
                    cluster_integrity = "false"
                current_node_cluster_size = node.get_statistic("cluster_size")
                if current_node_cluster_size != NOT_SUPPORTED:
                    if node_cluster_size is None:
                        node_cluster_size = current_node_cluster_size
                    elif node_cluster_size != current_node_cluster_size:
                        is_cluster_size_same = False

        if enterprise_flag and alerts_flag:
            self._alert.on_cluster_integrity_change(cluster_integrity)
            self._alert.on_cluster_size_change(is_cluster_size_same)
        new_node_list = node_list - self._node_list
        for node_address in new_node_list:
            actual_node_address = self.add_new_node(node_address)
            if actual_node_address is None:
                node_list.remove(node_address)
            elif enterprise_flag and alerts_flag:
                self._alert.on_new_node_add(actual_node_address)
        self.calculate_visibilities()

    def add_existing_node(self, node_address):

        ip, port = node_address.split(":")
        ip = ip_checker.check_ip(ip)

        node_address = "%s:%s" % (ip, port)

        for address in self._node_list:
            if address.startswith(ip + ":"):
                return dict(status="failure", error="Node already exists")

        if ip == "Unable to connect":
            return dict(status="failure", error="Unable to connect")
        elif ip == "Unable to resolve hostname":
            return dict(status="failure", error="Unable to resolve hostname")
        elif ip == "Please enter network IP":
            return dict(status="failure", error="Please enter network IP")
        else:
            self._node_list.add(node_address)
            self._nodes[node_address] = Node(node_address, self._cluster_id, validity=None)
            if self._nodes[node_address].get_status() == "on" and self.get_cluster_key() != self._nodes[node_address].get_cluster_key():
                self.remove_node(node_address)
                return dict(status="Node <strong>%s</strong> cannot be monitored here as it belongs to a different cluster." % node_address)
            else:
                return dict(status="success")

    def add_new_node(self, node_address):

        ip, port = node_address.split(":")
        node_address = base.get_self_address(ip, port)
        if node_address == UNABLE_TO_CONNECT or node_address in self._node_list:
            return None
        self._node_list.add(node_address)
        self._nodes[node_address] = Node(node_address, self._cluster_id)
        return node_address

    def remove_node(self, node_address):

        if node_address in self._node_list and node_address in self._nodes:
            self._node_list.remove(node_address)
            del self._nodes[node_address]
            return dict(status="success")
        return dict(status="failure")

    def get_id(self):
        return self._id

    def get_seed_ip(self):
        return self._ip

    def get_seed_port(self):
        return self._port

    def get_seed(self):
        return self._seed

    def set_seed(self, seed):
        self._seed = seed
        self._ip, self._port = seed.split(":")

    def get_status(self):
        return self._status

    def set_status_off(self):
        self._status = "off"

    def set_status_on(self):
        self._status = "on"

    def get_last_access_time(self):
        return self._last_access_time

    def update_last_access_time(self):
        self._last_access_time = datetime.now()

    def get_nodes(self):

        return self._nodes

    @HandleSeedDown()
    def get_build_no(self, ip=None, port=None):

        ip = ip or self._ip
        port = port or self._port
        return base.get_node_build(self._ip, self._port)

    def get_build_no_of_cluster_nodes(self):

        info = dict(version_list=dict())
        for node in list(self._nodes.values()):
            build_no = node.get_build()
            if build_no is not None:
                if build_no in list(info['version_list'].keys()):
                    info['version_list'][build_no].append(node.get_address())
                else:
                    info['version_list'][build_no] = [node.get_address()]
        build_list = []
        for build_no in list(info['version_list'].keys()):
            build_list.append(build_no.replace('.', ''))
        if list(info['version_list'].keys()) == []:
            info['latest_build_no'] = "Undefined"
        else:
            info['latest_build_no'] = list(info['version_list'].keys())[build_list.index(max(build_list))]
        return info

    def get_latest_build_no(self):

        return self._latest_build_no

    def get_node_list(self):

        return list(self._node_list)

    def get_off_nodes(self):

        off_nodes = []
        for node in list(self._nodes.values()):
            if node.get_status() == "off":
                off_nodes.append(node.get_address())
        return off_nodes

    def get_different_cluster_nodes(self):

        different_cluster_nodes = []
        for node in list(self._nodes.values()):
            if node.is_valid() is False:
                different_cluster_nodes.append(node.get_address())
                self.remove_node(node.get_address())
        return different_cluster_nodes

    def get_number_of_nodes(self):

        return len(self._node_list)

    def get_namespaces(self):

        self._namespaces = self.update_namespaces()
        return self._namespaces

    def get_number_of_namespaces(self):

        return len(self.get_namespaces())

    def update_namespaces(self):

        namespace_list = set()
        for node in list(self._nodes.values()):
            namespace_list = namespace_list.union(node.get_namespace_list())
        return list(namespace_list)

    def get_memory_usage(self):

        total_bytes_memory = 0
        used_bytes_memory = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_memory()
            if used.isdigit() and total.isdigit():
                total_bytes_memory += int(total)
                used_bytes_memory += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_memory = total_bytes_memory - used_bytes_memory

        return dict(used=used_bytes_memory, free=total_free_memory, details=details)

    def get_disk_usage(self):

        total_bytes_disk = 0
        used_bytes_disk = 0
        details = dict()

        for node_obj in list(self._nodes.values()):
            if node_obj is None or not node_obj.is_valid():
                continue

            used, total = node_obj.get_disk()
            if used.isdigit() and total.isdigit():
                total_bytes_disk += int(total)
                used_bytes_disk += int(used)
                details[node_obj.get_address()] = dict(used=int(used), free=int(total)-int(used))

        total_free_disk = total_bytes_disk - used_bytes_disk

        return dict(used=used_bytes_disk, free=total_free_disk, details=details)

    def get_throughput(self, node_addresses):

        throughput = dict()
        for node_address in node_addresses:
            node = self.get_node(node_address)
            if node is None or node.get_status() == "off" or not node_obj.is_valid():
                continue
            throughput[node_address] = node.get_throughput()
        return throughput

    def get_statistics(self):

        self._statistics = []
        for node in list(self._nodes.values()):
            if node.is_valid():
                stat = dict(address=node.get_address())
                stat.update(node.get_updated_statistics(self))
                self._statistics.append(stat)
        return self._statistics

    def update_statistics(self):

        for node in list(self._nodes.values()):
            if node.is_valid() is not False:
                node.get_updated_statistics(self)

    def calculate_visibilities(self):

        for node_obj in list(self._nodes.values()):
            if not node_obj.is_valid():
                continue
            nodes = node_obj.get_node_list()

            node_status = False if nodes == set() else True

            old_visibility = node_obj.get_cluster_visibility()
            visibility = True

            for node2_obj in list(self._nodes.values()):
                if not node2_obj.is_valid():
                    continue
                nodes2 = node2_obj.get_node_list()
                if node_obj.get_address() == node2_obj.get_address() or nodes2 == set():
                    continue

                if node_obj.get_address() in nodes2:
                    if nodes != set():
                        visibility &= True
                    else:
                        node_status = True
                elif node_obj.get_address() not in nodes2:
                    if nodes == set():
                        node_status |= False
                    else:
                        visibility = False
                        break

            if old_visibility != visibility:
                node_obj.set_cluster_visibility(visibility)
            if node_status == False:
                _off_nodes.add(node_obj.get_address())
                node_obj.set_status("off")
            else:
                node_obj.set_status("on")

    def get_node_visibility(self, node_address):

        return self._node_visibilties.get(node_address, None)

    def get_node(self, node_address):

        return self._nodes.get(node_address, None)

    def get_namespace(self, namespace_name, node_address):

        node_obj = self.get_node(node_address)
        if node_obj is None or not node_obj.is_valid():
            return None
        return node_obj.get_namespace(namespace_name)

    def get_xdr(self, xdr_port, node_address):

        node_obj = self.get_node(node_address)
        if node_obj is None or not node_obj.is_valid():
            return None
        return node_obj.get_xdr(xdr_port)

    def get_dc_names(self, xdr_port="3004"):

        dc_names = dict()
        for node in list(self._nodes.values()):
            address = node.get_address()
            xdr = self.get_xdr(xdr_port, address)
            if xdr is None:
                dc_names[address] = "node off"
            elif xdr == {}:
                dc_names[address] = "xdr off"
            else:
                dc_names[address] = xdr.get_dc_names()
        return dc_names

    @HandleSeedDownWithVersion()
    def get_indexes(self, ip=None, port=None):

        ip = ip or self._ip
        port = port or self._port
        try:
            result = base.get_indexes(ip, port)
        except:
            result = UNSUPPORTED_VERSION
        return result

    def get_indexes_by_namespaces(self, namespace):

        indexes = self.get_indexes()
        if indexes == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT
        elif indexes == UNSUPPORTED_VERSION:
            indexes = {}

        return indexes.get(namespace, [])

    def get_index_list_by_namespace(self):

        index_details = self.get_indexes()
        if index_details == UNABLE_TO_CONNECT:
            return {}
        elif index_details == UNSUPPORTED_VERSION:
            index_details = {}

        index_list = {}
        for namespace, indexes in list(index_details.items()):
            index_list[namespace] = [index["indexname"] for index in indexes]
        return index_list

    @HandleSeedDown()
    def get_sets_info(self, ip=None, port=None):

        ip = ip or self._ip
        port = port or self._port
        return base.get_sets_info(ip, port)

    def get_sets_info_by_namespaces(self, namespace):

        sets_info = self.get_sets_info()
        if sets_info == UNABLE_TO_CONNECT:
            return UNABLE_TO_CONNECT

        return sets_info.get(namespace, [])

    @HandleSeedDown()
    def get_cluster_key(self, ip=None, port=None):

        self._cluster_key = self._nodes[self._seed].get_cluster_key()
        return self._cluster_key
